/*   
     ----------------------------------------------------
     ----------  MAPSIM V0.1 ----------------------------
     ----------------------------------------------------
     
     Given a set of cosmic microwave background (CMB) anisotropy
     maps, T, Q, and U, and a pixelization, will compute
     the spherical harmonic mulitpoles from the maps.
     The input file has the map

     The input file has the following format:

     ----------------------------------------
     for each pixel i:

         T[i]    Q[i]    U[i]
     ----------------------------------------
     
     The pixel file must be named pixels.dat and
     has the following format
     ----------------------------------------
     for each pixel i:

         ra[i]    dec[i]
     ----------------------------------------

     where ra[i] and dec[i] are the ra and dec for pixel i given
     in degrees for both.  To run type:

      a.out 'filename' 'npix' 'lmax' 'seed'

     where 'filename' is the file containing the maps in the 
     format given above, 'npix' is the number of pixels in the file 
     pixels.dat, and 'lmax' is the maximum spherical harmonic multipole
     number.  

      a.out map.dat 65536 1500 -1

     Note that map.dat is the default filename of output
     from mapsim.

     M.E. Abroe    July 2002 
     University of Minnesota
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "nrutil.h"
#include "mapsim.h"
#include "mpi.h"

#define ME 1e-30

main(int argc,char **argv)
{
  FILE *f,*fmodel,*fmap;
  
  double **W,**X,*p,**C;
  double **M,**Y_p,**Y_n;
  double **at_p,**at_n;
  double **ae_p,**ae_n;
  double **ab_p,**ab_n;
  double *ra,*dec;
  double *x,*y,*z;
  double *Ct,*Ce,*Cb,*Cte,*Ctb,*Ceb;
  double sqct,sqce,sqcb,l11,l21,l22;
  double *T,*Q,*U;  
  double *cosine,*sine;  
  double mt=0,mq=0,mu=0,di;
  double x1,x2,x3;
  double pq,nq,pu,nu,lfac;

  int l,m,ni,my_n,my_s;  
  int i,j;
  int npix,lmax;
  int no_pe,my_pe;

  long idum;
  
  if(argc<4){
    printf("ERROR: Incorrect command line\n");
    exit(-1);}
  npix=atoi(argv[2]);
  lmax=atoi(argv[3]);

  /*********************/
  /* memory allocation */
  /*********************/
  ra  = dvector(0,npix-1);
  dec = dvector(0,npix-1);
  T   = dvector(0,npix-1);
  Q   = dvector(0,npix-1);
  U   = dvector(0,npix-1);
  x   = dvector(0,npix-1);
  y   = dvector(0,npix-1);
  z   = dvector(0,npix-1);
    
  sine   = dvector(0,lmax);
  cosine = dvector(0,lmax);
   
  M    = dmatrix(0,lmax,0,lmax);
  W    = dmatrix(0,lmax,0,lmax);
  X    = dmatrix(0,lmax,0,lmax);
  Y_p  = dmatrix(0,lmax,0,lmax);
  Y_n  = dmatrix(0,lmax,0,lmax);
  Ct   = dvector(0,lmax-1);
  Ce   = dvector(0,lmax-1);
  Cb   = dvector(0,lmax-1);
  Cte  = dvector(0,lmax-1);
  Ctb  = dvector(0,lmax-1);
  Ceb  = dvector(0,lmax-1);
      
  /******************/
  /* read in pixels */  
  /******************/
  if((f=fopen("pixels.dat","r"))==NULL){
    printf("ERROR:  no pixels.dat file\n");
    exit(-1);} i=0;
  while(!feof(f)){
    fscanf(f,"%lf%lf",&ra[i],&dec[i]);
    i++;}
  if((i-1)!=npix){
    printf("ERROR:  Incorrect number of pixels\n");
    exit(-1);}
     
  /* convert to spherical coordinates */
  rdtoxyz(ra,dec,x,y,z,npix);
    
  /* read in maps */
  if((fmodel=fopen(argv[1],"r"))==NULL){
    printf("ERROR:  no map file\n");
    exit(-1);}
  for(i=0 ; i<npix ; i++)
    fscanf(fmodel,"%lf%lf%lf",&T[i],&Q[i],&U[i]);
  
  /*****************/
  /* Initialze MPI */
  /*****************/
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &no_pe);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_pe);
  
  /* set individual proc index range */
  my_n=npix/no_pe;
  if(my_pe==no_pe-1) my_n+=npix%no_pe;
  my_s=my_pe*(npix/no_pe);
  printf("PE:%i assigned pixels %i to %i\n",my_pe,my_s,my_s+my_n-1);
  
  /* do the same for cl estimaties */
  for(l=2 ; l<=lmax ; l++){
    Ct[l]  = 0.0;
    Ce[l]  = 0.0;
    Cb[l]  = 0.0;
    Cte[l] = 0.0;
    Ctb[l] = 0.0;
    Ceb[l] = 0.0;}

  /********************/
  /* loop over pixels */
  /********************/
  for(i=my_s ; i<my_s+my_n ; i++){
   
    /* calculate relevant quantities */
    trig(x[i],y[i],cosine,sine,lmax);
    M_gen(z[i],M,lmax);
    Y_gen(M,Y_p,Y_n,cosine,sine,lmax);
    W_gen(z[i],M,W,lmax);
    X_gen(z[i],M,X,lmax);
    
    /* get T */
    for(l=2 ; l<=lmax ; l++){

      Ct[l] += (T[i]*Y_p[l][0])*(T[i]*Y_p[l][0]);
      
      for(m=1 ; m<=l ; m++){
	Ct[l] += (T[i]*Y_p[l][m])*(T[i]*Y_p[l][m]);
	Ct[l] += (T[i]*Y_n[l][m])*(T[i]*Y_n[l][m]);}
      
      
      
#if defined(PRINT)
      printf(" %i %i %lf %lf\n",my_pe,l,Ct[l],Y_p[l][m-1]); 
#endif
      
    }
    
  }
  
  /***********************/
  /* Gather maps to root */
  /***********************/

  printf("l, Ct: %i %lf\n",2,Ct[2]);

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Reduce(Ct,Ct,lmax+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  MPI_Reduce(Ce,Ce,lmax+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  MPI_Reduce(Cb,Cb,lmax+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  MPI_Reduce(Cte,Cte,lmax+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  MPI_Reduce(Ctb,Ctb,lmax+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  MPI_Reduce(Ceb,Ceb,lmax+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  
  printf("l, Ct: %i %lf\n",2,Ct[2]);
  
  /**********************/
  /* write maps to file */
  /**********************/  
  if(my_pe==0){
    fmap=fopen("cls.dat","w");
    for(l=0 ; l<lmax-1 ; l++){
      lfac=(double)(2.0*(double)(l)+1.0);
      fprintf(fmap,"%i %lf %lf %lf %lf %lf %lf\n",
	      l,Ct[l]/lfac,Cte[l]/lfac,Ctb[l]/lfac,
	      Ce[l]/lfac,Cb[l]/lfac,Ceb[l]/lfac);
    
      printf("%i %lf %lf %lf %lf %lf %lf :: %lf\n",
	     l,Ct[l]/lfac,Cte[l]/lfac,Ctb[l]/lfac,
	     Ce[l]/lfac,Cb[l]/lfac,Ceb[l]/lfac,lfac);
      
      
    }
    fclose(fmap);
  }  

  /*****************/
  /* Finailize MPI */
  /*****************/
  MPI_Finalize();
  return;
}







